package flink.zanxiangnet.ad.monitoring.process;

import flink.zanxiangnet.ad.monitoring.config.ClickhouseDataSourceFactory;
import flink.zanxiangnet.ad.monitoring.dao.mapper.AdStatOfDayDWDMapper;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdDataOfDayODS;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfDayDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.windows.GlobalWindow;
import org.apache.flink.util.Collector;
import org.apache.ibatis.datasource.DataSourceFactory;
import org.apache.ibatis.mapping.Environment;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.ibatis.transaction.jdbc.JdbcTransactionFactory;

import java.time.LocalDate;
import java.util.Map;
import java.util.Properties;
import java.util.stream.Collectors;

@Slf4j
public class AdDayDWDRollMonthProcess extends ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow> {
    private SqlSessionFactory sqlSessionFactory;
    // 上次查询的时间
    private ValueState<String> lastQueryDayState;
    // 之前聚合的昨天的数据
    private MapState<String, AdStatOfDayDWD> historyReduceState;

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();

        Properties ckProps = new Properties();
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_URL, params.get(ApplicationProperties.CK_URL));
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_USER, params.get(ApplicationProperties.CK_USERNAME));
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_PASSWORD, params.get(ApplicationProperties.CK_PASSWORD));

        DataSourceFactory dataSourceFactory = new ClickhouseDataSourceFactory();
        dataSourceFactory.setProperties(ckProps);
        Environment environment = new Environment("clickhouse", new JdbcTransactionFactory(), dataSourceFactory.getDataSource());
        org.apache.ibatis.session.Configuration configuration = new org.apache.ibatis.session.Configuration(environment);
        // 开启驼峰规则
        configuration.setMapUnderscoreToCamelCase(true);
        configuration.getTypeAliasRegistry().registerAlias(AdStatOfDayDWD.class);
        // addMapper一定要放到 alias的后面！！！！！
        configuration.addMapper(AdStatOfDayDWDMapper.class);
        sqlSessionFactory = new SqlSessionFactoryBuilder().build(configuration);

        lastQueryDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryDayState", String.class));
        historyReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("historyReduceState", String.class, AdStatOfDayDWD.class));
    }

    @Override
    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>.Context context,
                        Iterable<AdDataOfDayODS> iterable, Collector<AdStatOfDayDWD> collector) throws Exception {
        AdDataOfDayODS element = iterable.iterator().next();
        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
        long now = context.currentProcessingTime();
        long adId = element.getAdId();

        String lastQueryDay = lastQueryDayState.value();
        // 从 maxCompute查找广告的历史数据
        if (lastQueryDay == null || !lastQueryDay.equals(DateUtil.formatLocalDate(LocalDate.now()))) {
            try (SqlSession session = sqlSessionFactory.openSession()) {
                AdStatOfDayDWDMapper mapper = session.getMapper(AdStatOfDayDWDMapper.class);
                Map<String, AdStatOfDayDWD> historyData = mapper.lastReduceResult(adId, null, DateUtil.formatLocalDate(LocalDate.now().minusDays(1L)), 60).stream()
                        .sorted((o1, o2) -> new Long(o1.getCreateTime().getTime() - o2.getCreateTime().getTime()).intValue())
                        .collect(Collectors.toMap(AdStatOfDayDWD::getStatDay, data -> data, (val1, val2) -> val2));
                historyReduceState.clear();
                historyReduceState.putAll(historyData);
                lastQueryDayState.update(DateUtil.formatLocalDate(LocalDate.now()));
            }
        }
        AdStatOfDayDWD lastReduceData = null;
        for (int i = 1; i <= 60; i++) {
            lastReduceData = historyReduceState.get(DateUtil.formatLocalDate(statDay.minusDays(i)));
            if (lastReduceData != null) {
                break;
            }
        }
        AdStatOfDayDWD newStatData = AdStatOfDayDWD.reduce(lastReduceData, element, now);
        historyReduceState.put(DateUtil.formatLocalDate(statDay), newStatData);
        collector.collect(newStatData);
    }
}
